@@ -38,7 +38,7 @@ namespace Beagle.Daemon {
 		public delegate void StartedHandler (QueryResult source);
 		public event StartedHandler StartedEvent;
 
-		public delegate void HitsAddedHandler (QueryResult source, ICollection someHits);
+		public delegate void HitsAddedHandler (QueryResult source, ICollection someHits, int total_results);
 		public event HitsAddedHandler HitsAddedEvent;
 
 		public delegate void HitsSubtractedHandler (QueryResult source, ICollection someUris);
@@ -98,9 +98,14 @@ namespace Beagle.Daemon {
 			}
 		}
 
+		public void Add (ICollection some_hits)
+		{
+			Add (some_hits, -1);
+		}
+
 		// Note: some_hits is allowed to contain null.
 		// They are silently ignored.
-		public void Add (ICollection some_hits)
+		public void Add (ICollection some_hits, int total_results)
 		{
 			lock (this) {
 				if (cancelled)
@@ -113,7 +118,7 @@ namespace Beagle.Daemon {
 
 				if (IsIndexListener) {
 					if (HitsAddedEvent != null)
-						HitsAddedEvent (this, some_hits);
+						HitsAddedEvent (this, some_hits, total_results);
 					return;
 				}
 
@@ -128,7 +133,7 @@ namespace Beagle.Daemon {
 				}
 				
 				if (HitsAddedEvent != null && hits_to_report.Count > 0)
-					HitsAddedEvent (this, hits_to_report);
+					HitsAddedEvent (this, hits_to_report, total_results);
 			}
 		}
 
